@@ -249,8 +249,16 @@ class RestResponse(object):
     def dict(self):
         """Property for accessing the data as an dict"""
         try:
+            if len(self.text) == 0:
+                # No response body; return empty dict instead to avoid exceptions
+                # No response bodies can be valid in many cases (especially 4XX and 5XX responses)
+                return {}
             return json.loads(self.text)
         except:
+            if self.status == 500:
+                # Make an allowance for 500 status codes
+                # Depending on the reason for the error, it's possible the web server may not be able to support Redfish handling
+                return {}
             str = "Service responded with invalid JSON at URI {}\n{}".format(
                 self._rest_request.path, self.text)
             LOGGER.error(str)
@@ -280,8 +288,14 @@ class RestResponse(object):
         """Property for accessing the saved session location"""
         if self._session_location:
             return self._session_location
+        if self.status not in [200, 201, 202, 204]:
+            return None
 
         self._session_location = self.getheader('location')
+        if self._session_location is None:
+            warnings.warn("Service incorrectly did not provide the 'Location' response header when creating a session; attempting to "
+                          "get the session location from the response body.  Contact your vendor.")
+            self._session_location = self.dict["@odata.id"]
         return self._session_location
 
     @property
@@ -454,7 +468,8 @@ class RestClientBase(object):
     def __init__(self, base_url, username=None, password=None,
                                 default_prefix='/redfish/v1/', sessionkey=None,
                                 capath=None, cafile=None, timeout=None,
-                                max_retry=None, proxies=None, check_connectivity=True):
+                                max_retry=None, proxies=None, check_connectivity=True,
+                                https_adapter = None):
         """Initialization of the base class RestClientBase
 
         :param base_url: The URL of the remote system
@@ -480,7 +495,7 @@ class RestClientBase(object):
         :param check_connectivity: A boolean to determine whether the client immediately checks for
         connectivity to the base_url or not.
         :type check_connectivity: bool
-
+        :type https_adapter: requests.adpaters.HTTPAdapter
         """
 
         self.__base_url = base_url.rstrip('/')
@@ -493,6 +508,8 @@ class RestClientBase(object):
             self._session = requests_unixsocket.Session()
         else:
             self._session = requests.Session()
+            if https_adapter:
+                self._session.mount('https://',https_adapter)
         self._timeout = timeout
         self._max_retry = max_retry if max_retry is not None else 10
         self._proxies = proxies
@@ -603,8 +620,11 @@ class RestClientBase(object):
             self.root = {}
             self.root_resp = resp
             return
+        if resp.status == 503:
+            raise ServerDownOrUnreachableError("Service is busy, " \
+                                               "return code: %d" % resp.status,response=resp)
         if resp.status != 200:
-            raise ServerDownOrUnreachableError("Server not reachable, " \
+            raise ServerDownOrUnreachableError("Service not able to provide the service root, " \
                                                "return code: %d" % resp.status,response=resp)
 
         content = resp.text
@@ -969,7 +989,7 @@ class RestClientBase(object):
         else:
             raise RetriesExhaustedError() from cause_exception
 
-    def login(self, username=None, password=None, auth=AuthMethod.SESSION):
+    def login(self, username=None, password=None, auth=AuthMethod.SESSION, headers=None):
         """Login and start a REST session.  Remember to call logout() when"""
         """ you are done.
 
@@ -979,6 +999,8 @@ class RestClientBase(object):
         :type password: str.
         :param auth: authentication method
         :type auth: object/instance of class AuthMethod
+        :param headers: Additional HTTP headers to provide in the request
+        :type headers: dict, optional
 
         """
         if getattr(self, "root_resp", None) is None:
@@ -987,12 +1009,13 @@ class RestClientBase(object):
         self.__username = username if username else self.__username
         self.__password = password if password else self.__password
 
+        headers = headers if headers else {}
+
         if auth == AuthMethod.BASIC:
             auth_key = base64.b64encode(('%s:%s' % (self.__username,
                             self.__password)).encode('utf-8')).decode('utf-8')
             self.__authorization_key = 'Basic %s' % auth_key
 
-            headers = dict()
             headers['Authorization'] = self.__authorization_key
 
             respvalidate = self._rest_request(self.login_url, headers=headers)
@@ -1005,7 +1028,6 @@ class RestClientBase(object):
             data['UserName'] = self.__username
             data['Password'] = self.__password
 
-            headers = dict()
             resp = self._rest_request(self.login_url, method="POST",body=data,
                                       headers=headers, allow_redirects=False)
 
@@ -1032,7 +1054,7 @@ class RestClientBase(object):
                             error_str = resp.dict["error"]["message"]
                         except:
                             pass
-                    raise SessionCreationError('HTTP {}: Failed to created the session\n{}'.format(resp.status, error_str))
+                    raise SessionCreationError('HTTP {}: Failed to create the session\n{}'.format(resp.status, error_str))
         else:
             pass
 
@@ -1060,7 +1082,8 @@ class HttpClient(RestClientBase):
                                 default_prefix='/redfish/v1/',
                                 sessionkey=None, capath=None,
                                 cafile=None, timeout=None,
-                                max_retry=None, proxies=None, check_connectivity=True):
+                                max_retry=None, proxies=None, check_connectivity=True,
+                                https_adapter=None):
         """Initialize HttpClient
 
         :param base_url: The url of the remote system
@@ -1086,14 +1109,15 @@ class HttpClient(RestClientBase):
         :param check_connectivity: A boolean to determine whether the client immediately checks for
         connectivity to the base_url or not.
         :type check_connectivity: bool
-
+        :param https_adapter session adapter for HTTPS
+        :type https_adapter: requests.adpaters.HTTPAdapter
         """
         super(HttpClient, self).__init__(base_url, username=username,
                             password=password, default_prefix=default_prefix,
                             sessionkey=sessionkey, capath=capath,
                             cafile=cafile, timeout=timeout,
                             max_retry=max_retry, proxies=proxies,
-                            check_connectivity=check_connectivity)
+                            check_connectivity=check_connectivity,https_adapter=https_adapter)
 
         try:
             self.login_url = self.root.Links.Sessions['@odata.id']
@@ -1150,7 +1174,8 @@ def redfish_client(base_url=None, username=None, password=None,
                                 default_prefix='/redfish/v1/',
                                 sessionkey=None, capath=None,
                                 cafile=None, timeout=None,
-                                max_retry=None, proxies=None, check_connectivity=True):
+                                max_retry=None, proxies=None, check_connectivity=True,
+                                https_adapter=None):
     """Create and return appropriate REDFISH client instance."""
     """ Instantiates appropriate Redfish object based on existing"""
     """ configuration. Use this to retrieve a pre-configured Redfish object
@@ -1177,14 +1202,17 @@ def redfish_client(base_url=None, username=None, password=None,
     :type proxies: dict
     :param check_connectivity: A boolean to determine whether the client immediately checks for
     connectivity to the base_url or not.
-    :type check_connectivity: bool
+    :type check_connectivity: bo#ol
+    :param https_adapter session adapter for HTTPS
+    :type https_adapter: requests.adpaters.HTTPAdapter
     :returns: a client object.
 
     """
     if "://" not in base_url:
-        warnings.warn("Scheme not specified for '{}'; adding 'https://'".format(base_url))
-        base_url = "https://" + base_url
+            warnings.warn("Scheme not specified for '{}'; adding 'https://'".format(base_url))
+            base_url = "https://" + base_url
     return HttpClient(base_url=base_url, username=username, password=password,
                         default_prefix=default_prefix, sessionkey=sessionkey,
                         capath=capath, cafile=cafile, timeout=timeout,
-                        max_retry=max_retry, proxies=proxies, check_connectivity=check_connectivity)
+                        max_retry=max_retry, proxies=proxies, check_connectivity=check_connectivity,
+                        https_adapter=https_adapter)
